# "screenkey" is distributed under GNU GPLv3+, WITHOUT ANY WARRANTY.
# Copyright(c) 2010-2012: Pablo Seminario <pabluk@gmail.com>
# Copyright(c) 2015-2020: wave++ "Yuri D'Elia" <wavexx@thregr.org>
# Copyright(c) 2019-2020: Yuto Tokunaga <yuntan.sub1@gmail.com>

from . import *
from .labelmanager import LabelManager

from datetime import datetime
import json
import os
import subprocess
import numbers
from tempfile import NamedTemporaryFile

import gi
gi.require_version('Gtk', '3.0')
gi.require_version('Pango', '1.0')

from gi.repository import GLib, Gtk, Gdk, GdkPixbuf, Pango, GObject
import cairo


# Gtk shortcuts
START = Gtk.Align.START
CENTER = Gtk.Align.CENTER
END = Gtk.Align.END
FILL = Gtk.Align.FILL
TOP = Gtk.PositionType.TOP
BOTTOM = Gtk.PositionType.BOTTOM
RIGHT = Gtk.PositionType.RIGHT
LEFT = Gtk.PositionType.LEFT
HORIZONTAL = Gtk.Orientation.HORIZONTAL
VERTICAL = Gtk.Orientation.VERTICAL
IF_VALID = Gtk.SpinButtonUpdatePolicy.IF_VALID


BUTTONS_MIN_BLINK = 1/30        # Minimum persistence for any action (s)
BUTTONS_REL_BRIGHT = 127        # Residual brightness after button release


# SVG Data for mouse buttons
BUTTONS_SVG = None

def load_button_pixbufs(color):
    global BUTTONS_SVG

    if BUTTONS_SVG is None:
        image_path = os.path.join(MODULE_DIR, 'images', 'mouse.svg')
        with open(image_path) as svg_file:
            BUTTONS_SVG = svg_file.readlines()

    if not isinstance(color, str):
        # Gdk.Color
        color = 'rgb({}, {}, {})'.format(
            round(color.red_float * 255),
            round(color.green_float * 255),
            round(color.blue_float * 255)
        )
    button_pixbufs = []
    svg = NamedTemporaryFile(mode='w', suffix='.svg')
    for line in BUTTONS_SVG[1:-1]:
        svg.seek(0)
        svg.truncate()
        svg.writelines((
            BUTTONS_SVG[0],
            line.replace('#fff', color),
            BUTTONS_SVG[-1],
        ))
        svg.flush()
        os.fsync(svg.fileno())
        button_pixbufs.append(GdkPixbuf.Pixbuf.new_from_file(svg.name))
    svg.close()
    return button_pixbufs


def gi_module_available(module, version):
    try:
        gi.require_version(module, version)
        return True
    except ValueError:
        return False


class Screenkey(Gtk.Window):
    STATE_FILE = os.path.join(GLib.get_user_config_dir(), 'screenkey.json')

    def __init__(self, logger, options, show_settings=False):
        self.logger = logger
        self.logger.debug("{} {}".format(APP_NAME, VERSION))

        self.exit_status = None
        self.timer_hide = None
        self.timer_min = None

        defaults = Options({'no_systray': False,
                            'timeout': 2.5,
                            'recent_thr': 0.1,
                            'compr_cnt': 3,
                            'ignore': [],
                            'position': 'bottom',
                            'persist': False,
                            'window': False,
                            'font_desc': 'Sans Bold',
                            'font_size': 'medium',
                            'font_color': 'white',
                            'bg_color': 'black',
                            'opacity': 0.8,
                            'key_mode': 'composed',
                            'bak_mode': 'baked',
                            'mods_mode': 'normal',
                            'mods_only': False,
                            'multiline': False,
                            'vis_shift': False,
                            'vis_space': True,
                            'geometry': None,
                            'screen': 0,
                            'start_disabled': False,
                            'mouse': False,
                            'button_hide_duration': 1})
        self.options = self.load_state()
        if self.options is None:
            self.options = defaults
        else:
            # copy missing defaults
            for k, v in defaults.items():
                if k not in self.options:
                    self.options[k] = v
        if options is not None:
            # override with values from constructor
            for k, v in options.items():
                if v is not None:
                    self.options[k] = v


        if not self.options.window:
            Gtk.Window.__init__(self, Gtk.WindowType.POPUP)
        else:
            self.options.persist = True
            Gtk.Window.__init__(self, Gtk.WindowType.TOPLEVEL)

        self.set_keep_above(True)
        self.set_accept_focus(False)
        self.set_focus_on_map(False)
        self.set_app_paintable(True)

        self.button_pixbufs = []
        self.button_states = [None] * 11
        self.img = Gtk.Image()
        self.update_image_tag = None

        self.box = Gtk.HBox(homogeneous=False)
        self.box.show()
        self.add(self.box)

        self.label = Gtk.Label()
        self.label.set_ellipsize(Pango.EllipsizeMode.START)
        self.label.set_justify(Gtk.Justification.CENTER)
        self.label.show()

        self.font = Pango.FontDescription(self.options.font_desc)
        self.update_colors()
        self.update_mouse_enabled()

        self.set_size_request(0, 0)
        self.set_gravity(Gdk.Gravity.CENTER)
        self.connect("configure-event", self.on_configure)
        self.connect("draw", self.on_draw)

        scr = self.get_screen()
        scr.connect("size-changed", self.on_screen_size_changed)
        scr.connect("monitors-changed", self.on_monitors_changed)
        self.set_active_monitor(self.options.screen)

        visual = scr.get_rgba_visual()
        if visual is not None:
            self.set_visual(visual)

        self.box.pack_start(self.img, expand=False, fill=True, padding=0)
        self.box.pack_end(self.label, expand=True, fill=True, padding=0)

        self.labelmngr = None
        self.enabled = True
        self.on_change_mode()

        self.make_menu()
        self.make_about_dialog()
        self.make_preferences_dialog()

        if not self.options.no_systray:
            if gi_module_available('AppIndicator3', '0.1'):
                self.make_appindicator()
            else:
                self.make_systray()

        self.connect("delete-event", self.quit)
        if show_settings:
            self.on_preferences_dialog()
        if self.options.persist:
            self.show()


    def quit(self, widget=None, data=None, exit_status=os.EX_OK):
        self.labelmngr.stop()
        self.exit_status = exit_status
        Gtk.main_quit()


    def load_state(self):
        """Load stored options"""
        options = None
        try:
            with open(self.STATE_FILE) as f:
                options = Options(json.load(f))
                self.logger.debug("Options loaded.")
        except OSError:
            self.logger.debug("file %s does not exists." % self.STATE_FILE)
        except ValueError:
            self.logger.debug("file %s is invalid." % self.STATE_FILE)

        # compatibility with previous versions (0.5)
        if options and options.key_mode == 'normal':
            options.key_mode = 'composed'

        return options


    def store_state(self, options):
        """Store options"""
        try:
            with open(self.STATE_FILE, 'w') as f:
                json.dump(options._store, f, indent=4)
                self.logger.debug("Options saved.")
        except OSError:
            self.logger.debug("Cannot open %s." % self.STATE_FILE)


    def set_active_monitor(self, monitor):
        scr = self.get_screen()
        if monitor >= scr.get_n_monitors():
            self.monitor = 0
        else:
            self.monitor = monitor
        self.update_geometry()


    def on_monitors_changed(self, monitor):
        self.set_active_monitor(self.monitor)


    def update_mouse_enabled(self):
        if self.options.mouse:
            if not self.button_pixbufs:
                self.button_pixbufs = load_button_pixbufs(
                    Gdk.color_parse(self.options.font_color)
                )
            self.img.show()
            self.update_image_tag = GLib.idle_add(self.update_image)
        else:
            self.img.hide()
            if self.update_image_tag is not None:
                GLib.source_remove(self.update_image_tag)
                self.update_image_tag = None


    def do_get_preferred_height(self):
        return self.height


    def update_font(self):
        text = self.label.get_text()
        lines = text.count('\n') + 1
        self.font.set_absolute_size((50 * self.height // lines // 100) * 1000)
        self.label.set_padding(self.width // 100, 0)
        self.label.get_pango_context().set_font_description(self.font)


    def update_image(self):
        if not self.button_pixbufs:
            self.update_image_tag = None
            return False

        pixbuf = self.button_pixbufs[0]
        copied = False

        for index, button_state in enumerate(self.button_states):
            if button_state is None:
                continue
            delta_time = (datetime.now() - button_state.stamp).total_seconds()
            if button_state.pressed or delta_time < BUTTONS_MIN_BLINK:
                alpha = 255
            elif self.options.button_hide_duration > 0:
                hide_time = delta_time / self.options.button_hide_duration
                alpha = int(BUTTONS_REL_BRIGHT * (1 - min(1, hide_time)))
            else:
                alpha = 0

            if not copied:
                pixbuf = pixbuf.copy()
                copied = True
            self.button_pixbufs[button_state.btn].composite(
                pixbuf, 0, 0, pixbuf.get_width(), pixbuf.get_height(),
                0, 0, 1, 1,
                GdkPixbuf.InterpType.NEAREST, alpha)

            if not button_state.pressed and \
               delta_time >= self.options.button_hide_duration:
                self.button_states[index] = None

        scale = self.height / pixbuf.get_height()
        if scale != 1:
            width = int(pixbuf.get_width() * scale)
            pixbuf = pixbuf.scale_simple(width, self.height, GdkPixbuf.InterpType.BILINEAR)
        self.img.set_from_pixbuf(pixbuf)

        if not copied:
            self.update_image_tag = None
            return False
        return True


    def update_colors(self):
        font_color = Gdk.color_parse(self.options.font_color)
        self.label.modify_fg(Gtk.StateFlags.NORMAL, font_color)
        self.bg_color = Gdk.color_parse(self.options.bg_color)
        if self.options.mouse and self.button_pixbufs:
            self.button_pixbufs = load_button_pixbufs(font_color)
        self.queue_draw()


    def on_draw(self, widget, cr):
        cr.set_source_rgba(self.bg_color.red_float,
                           self.bg_color.green_float,
                           self.bg_color.blue_float,
                           self.options.opacity)
        cr.set_operator(cairo.OPERATOR_SOURCE)
        cr.paint()
        cr.set_operator(cairo.OPERATOR_OVER)
        return False


    def on_configure(self, event, data):
        # set event mask for click-through
        self.input_shape_combine_region(cairo.Region(cairo.RectangleInt(0, 0, 0, 0)))


    def on_screen_size_changed(self, screen):
        self.width, self.height = self.get_size()
        self.update_font()
        self.update_image()


    def update_geometry(self, configure=False):
        geometry = self.get_screen().get_monitor_geometry(self.monitor)

        if self.options.geometry is not None:
            # NOTE: this assume a single global scaling factor for all
            # monitors which seems to be true for GTK3:
            # https://wiki.gnome.org/HowDoI/HiDpi
            scale = self.get_screen().get_monitor_scale_factor(self.monitor)

            # scale the coordinate according to the selected monitor
            # or convert from device to gdk pixel units
            def coord_to_gdk(coord, extent):
                if not isinstance(coord, numbers.Integral):
                    return int(coord * extent)
                else:
                    return int(coord / scale)

            x, y, w, h = self.options.geometry
            x = coord_to_gdk(x, geometry.width)
            y = coord_to_gdk(y, geometry.height)
            w = coord_to_gdk(w, geometry.width)
            h = coord_to_gdk(h, geometry.height)
            if x < 0: x = geometry.width + x - w
            if y < 0: y = geometry.height + y - h
            area_geometry = [x, y, w, h]
        else:
            area_geometry = [geometry.x, geometry.y, geometry.width, geometry.height]

        if self.options.position == 'fixed':
            self.move_resize(*area_geometry)
            return

        if self.options.font_size == 'large':
            window_height = 24 * area_geometry[3] // 100
        elif self.options.font_size == 'medium':
            window_height = 12 * area_geometry[3] // 100
        else:
            window_height = 8 * area_geometry[3] // 100

        if self.options.position == 'top':
            window_y = area_geometry[1] + area_geometry[3] // 10
        elif self.options.position == 'center':
            window_y = area_geometry[1] + area_geometry[3] // 2 - window_height // 2
        else:
            window_y = area_geometry[1] + area_geometry[3] * 9 // 10 - window_height

        self.move_resize(area_geometry[0], window_y,
                         area_geometry[2], window_height)


    def move_resize(self, x, y, w, h):
        self.width = w
        self.height = h
        self.move(x, y)
        self.resize(w, h)


    def on_statusicon_popup(self, widget, button, timestamp, data=None):
        if button == 3 and data:
            data.show()
            data.popup_at_pointer(None)


    def on_labelmngr_error(self):
        msg = Gtk.MessageDialog(parent=self,
                                type=Gtk.MessageType.ERROR,
                                buttons=Gtk.ButtonsType.OK,
                                message_format="Error initializing Screenkey")
        text = _('Screenkey failed to initialize. This is usually a sign of an improperly '
                 'configured input method or desktop keyboard settings. Please see the <a '
                 'href="{url}">troubleshooting documentation</a> for further diagnosing '
                 'instructions.\n\nScreenkey cannot recover and will now quit!')
        msg.format_secondary_markup(text.format(url=ERROR_URL))
        msg.run()
        msg.destroy()
        self.quit(exit_status=os.EX_SOFTWARE)


    def timed_show(self):
        if not self.get_property('visible'):
            self.show()
        if self.timer_hide is not None:
            GObject.source_remove(self.timer_hide)
            self.timer_hide = None
        if self.options.timeout > 0:
            # hide automatically if mouse mode is disabled. keep the
            # window around otherwise as long as any of the visible keys
            # (mouse or modifiers) is still held
            if not self.options.mouse or \
               not any(b and b.pressed for b in self.button_states):
                self.timer_hide = GObject.timeout_add(self.options.timeout * 1000, self.on_timeout_main)


    def on_label_change(self, markup, synthetic):
        if markup is None:
            self.on_labelmngr_error()
            return

        _, attr, text, _ = Pango.parse_markup(markup, -1, '\0')
        self.label.set_text(text)
        self.label.set_attributes(attr)
        self.update_font()

        self.timed_show()
        if self.timer_min is not None:
            GObject.source_remove(self.timer_min)
            self.timer_min = None
        if not synthetic:
            self.timer_min = GObject.timeout_add(self.options.recent_thr * 2000, self.on_timeout_min)


    def on_image_change(self, button_state):
        if button_state:
            btn = button_state.btn
            # Don't do animation after stealth enable
            if self.button_states[btn] is not None or button_state.pressed:
                self.button_states[btn] = button_state
                if self.options.mouse:
                    if not self.update_image_tag:
                        self.update_image_tag = GLib.idle_add(self.update_image)
                    self.timed_show()
        else:
            # Reset all
            self.button_states = [None for _ in self.button_states]
            if self.options.mouse:
                if not self.update_image_tag:
                    self.update_image_tag = GLib.idle_add(self.update_image)
                self.timed_show()


    def on_timeout_main(self):
        if not self.options.persist:
            self.hide()
        self.timer_hide = None
        self.label.set_text('')
        self.labelmngr.clear()
        return False


    def on_timeout_min(self):
        self.timer_min = None
        self.labelmngr.queue_update()
        return False


    def restart_labelmanager(self):
        self.logger.debug("Restarting LabelManager.")
        if self.labelmngr:
            self.labelmngr.stop()
        self.labelmngr = LabelManager(self.on_label_change,
                                      self.on_image_change,
                                      logger=self.logger,
                                      key_mode=self.options.key_mode,
                                      bak_mode=self.options.bak_mode,
                                      mods_mode=self.options.mods_mode,
                                      mods_only=self.options.mods_only,
                                      multiline=self.options.multiline,
                                      vis_shift=self.options.vis_shift,
                                      vis_space=self.options.vis_space,
                                      recent_thr=self.options.recent_thr,
                                      compr_cnt=self.options.compr_cnt,
                                      ignore=self.options.ignore,
                                      pango_ctx=self.label.get_pango_context(),
                                      enabled=not self.options.start_disabled)
        self.labelmngr.start()


    def on_change_mode(self):
        if not self.enabled:
            return
        self.restart_labelmanager()


    def on_show_keys(self, widget, data=None):
        self.enabled = widget.get_active()
        if self.enabled:
            self.logger.debug("Screenkey enabled.")
            self.restart_labelmanager()
        else:
            self.logger.debug("Screenkey disabled.")
            self.labelmngr.stop()


    def on_preferences_dialog(self, widget=None, data=None):
        self.prefs.show()


    def on_preferences_changed(self, widget=None, data=None):
        self.store_state(self.options)
        self.prefs.hide()
        return True


    def make_preferences_dialog(self):
        # TODO: switch to something declarative or at least clean-up the following mess
        self.prefs = prefs = Gtk.Dialog(APP_NAME, None,
                                        Gtk.DialogFlags.DESTROY_WITH_PARENT,
                                        (Gtk.STOCK_CLOSE, Gtk.ResponseType.CLOSE),
                                        use_header_bar=True,
                                        destroy_with_parent=True,
                                        resizable=False)
        prefs.connect("response", self.on_preferences_changed)
        prefs.connect("delete-event", self.on_preferences_changed)

        def on_sb_time_changed(widget, data=None):
            self.options.timeout = widget.get_value()
            self.logger.debug("Timeout value changed: %f." % self.options.timeout)

        def on_cbox_sizes_changed(widget, data=None):
            self.options.font_size = widget.props.active_id
            self.update_geometry()
            self.logger.debug("Window size changed: %s." % self.options.font_size)

        def on_cbox_modes_changed(widget, data=None):
            self.options.key_mode = widget.props.active_id
            self.on_change_mode()
            self.logger.debug("Key mode changed: %s." % self.options.key_mode)

        def on_cbox_bak_changed(widget, data=None):
            self.options.bak_mode = widget.props.active_id
            self.on_change_mode()
            self.logger.debug("Bak mode changed: %s." % self.options.bak_mode)

        def on_cbox_mods_changed(widget, data=None):
            self.options.mods_mode = widget.props.active_id
            self.on_change_mode()
            self.logger.debug("Mods mode changed: %s." % self.options.mods_mode)

        def on_cbox_modsonly_changed(widget, data=None):
            self.options.mods_only = widget.get_active()
            self.on_change_mode()
            self.logger.debug("Modifiers only changed: %s." % self.options.mods_only)

        def on_cbox_visshift_changed(widget, data=None):
            self.options.vis_shift = widget.get_active()
            self.on_change_mode()
            self.logger.debug("Visible Shift changed: %s." % self.options.vis_shift)

        def on_cbox_visspace_changed(widget, data=None):
            self.options.vis_space = widget.get_active()
            self.on_change_mode()
            self.logger.debug("Show Whitespace changed: %s." % self.options.vis_space)

        def on_cbox_position_changed(widget, data=None):
            new_position = widget.props.active_id
            if self.options.position != 'fixed' and new_position == 'fixed':
                new_geom = on_btn_sel_geom(widget)
                if not new_geom:
                    self.cbox_positions.props.active_id = self.options.position
                    return
            self.options.position = new_position
            self.update_geometry()
            self.logger.debug("Window position changed: %s." % self.options.position)

        def on_cbox_screen_changed(widget, data=None):
            self.options.screen = widget.get_active()
            self.set_active_monitor(self.options.screen)
            self.logger.debug("Screen changed: %d." % self.options.screen)

        def on_cbox_persist_changed(widget, data=None):
            self.options.persist = widget.get_active()
            if not self.get_property('visible'):
                self.show()
            else:
                self.on_label_change(self.label.get_text(), True)
            self.logger.debug("Persistent changed: %s." % self.options.persist)

        def on_sb_compr_changed(widget, data=None):
            self.options.compr_cnt = widget.get_value_as_int()
            self.on_change_mode()
            self.logger.debug("Compress repeats value changed: %d." % self.options.compr_cnt)

        def on_cbox_compr_changed(widget, data=None):
            compr_enabled = widget.get_active()
            self.sb_compr.set_sensitive(compr_enabled)
            self.options.compr_cnt = self.sb_compr.get_value_as_int() if compr_enabled else 0
            self.on_change_mode()
            self.logger.debug("Compress repeats value changed: %d." % self.options.compr_cnt)

        def on_btn_sel_geom(widget, data=None):
            try:
                ret = subprocess.check_output(['slop', '-f', '%x %y %w %h %i'])
            except subprocess.CalledProcessError:
                return False
            except OSError:
                msg = Gtk.MessageDialog(parent=self,
                                        type=Gtk.MessageType.ERROR,
                                        buttons=Gtk.ButtonsType.OK,
                                        message_format="Error running \"slop\"")
                msg.format_secondary_markup(_("\"slop\" is required for interactive selection. "
                                              "See <a href=\"{url}\">{url}</a>").format(url=SLOP_URL))
                msg.run()
                msg.destroy()
                return False

            ret = ret.decode("utf8")
            data = list(map(int, ret.split(' ')))
            self.options.geometry = data[0:4]
            self.options.window = data[4]
            if not self.options.window or \
               self.options.window == self.get_screen().get_root_window().get_xid():
                # region selected, switch to fixed
                self.options.window = None
                self.options.position = 'fixed'
                self.cbox_positions.props.active_id = self.options.position

            self.update_geometry()
            self.btn_reset_geom.set_sensitive(True)
            return True

        def on_btn_reset_geom(widget, data=None):
            self.options.geometry = None
            if self.options.position == 'fixed':
                self.options.position = 'bottom'
                self.cbox_positions.props.active_id = self.options.position
            self.update_geometry()
            widget.set_sensitive(False)

        def on_adj_opacity_changed(widget, data=None):
            self.options.opacity = widget.get_value()
            self.update_colors()

        def on_font_color_changed(widget, data=None):
            self.options.font_color = widget.get_color().to_string()
            self.update_colors()

        def on_bg_color_changed(widget, data=None):
            self.options.bg_color = widget.get_color().to_string()
            self.update_colors()

        def on_btn_font(widget, data=None):
            widget.props.label = widget.props.font
            self.options.font_desc = widget.props.font
            self.font = widget.props.font_desc
            self.update_font()

        def on_cbox_mouse_changed(widget, data=None):
            self.options.mouse = widget.get_active()
            self.logger.debug("Mouse changed: %s." % self.options.mouse)
            self.update_mouse_enabled()

        def on_sb_mouse_duration_changed(widget, data=None):
            self.options.button_hide_duration = widget.get_value()
            self.logger.debug("Button hide duration value changed: %f." % self.options.button_hide_duration)

        frm_time = Gtk.Frame(label_widget=Gtk.Label("<b>%s</b>" % _("Time"),
                                                    use_markup=True),
                             border_width=4,
                             shadow_type=Gtk.ShadowType.NONE,
                             margin=6, hexpand=True)
        vbox_time = Gtk.Grid(orientation=VERTICAL,
                             row_spacing=6, margin=6)
        hbox_time = Gtk.Grid(column_spacing=6)
        lbl_time1 = Gtk.Label(_("Display for"))
        lbl_time2 = Gtk.Label(_("seconds"))
        sb_time = Gtk.SpinButton(digits=1,
                                 numeric=True,
                                 update_policy=IF_VALID)
        sb_time.set_increments(0.5, 1.0)
        sb_time.set_range(0, 300)
        sb_time.set_value(self.options.timeout)
        sb_time.connect("value-changed", on_sb_time_changed)
        hbox_time.add(lbl_time1)
        hbox_time.add(sb_time)
        hbox_time.add(lbl_time2)
        vbox_time.add(hbox_time)

        chk_persist = Gtk.CheckButton(_("Persistent window"),
                                      active=self.options.persist)
        chk_persist.connect("toggled", on_cbox_persist_changed)
        vbox_time.add(chk_persist)

        frm_time.add(vbox_time)

        frm_position = Gtk.Frame(label_widget=Gtk.Label("<b>%s</b>" % _("Position"),
                                                        use_markup=True),
                                 border_width=4,
                                 shadow_type=Gtk.ShadowType.NONE,
                                 margin=6, hexpand=True)
        grid_position = Gtk.Grid(row_spacing=6, column_spacing=6,
                                 margin=6)

        lbl_screen = Gtk.Label(_("Screen"),
                               halign=START)
        cbox_screen = Gtk.ComboBoxText()
        scr = self.get_screen()
        for i in range(scr.get_n_monitors()):
            cbox_screen.insert_text(i, '%d: %s' % (i, scr.get_monitor_plug_name(i)))
        cbox_screen.set_active(self.monitor)
        cbox_screen.connect("changed", on_cbox_screen_changed)

        lbl_positions = Gtk.Label(_("Position"),
                                  halign=START)
        self.cbox_positions = Gtk.ComboBoxText(name='position')
        for id_, text in POSITIONS.items():
            self.cbox_positions.append(id_, text)
            if id_ == self.options.position:
                self.cbox_positions.props.active_id = id_
        self.cbox_positions.connect("changed", on_cbox_position_changed)

        self.btn_reset_geom = Gtk.Button(_("Reset"))
        self.btn_reset_geom.connect("clicked", on_btn_reset_geom)
        self.btn_reset_geom.set_sensitive(self.options.geometry is not None)

        hbox_position = Gtk.Grid(column_spacing=6, halign=END)
        hbox_position.add(self.cbox_positions)
        hbox_position.add(self.btn_reset_geom)

        btn_sel_geom = Gtk.Button(_("Select window/region"),
                                  halign=FILL, hexpand=True)
        btn_sel_geom.connect("clicked", on_btn_sel_geom)

        grid_position.add(lbl_screen)
        grid_position.attach_next_to(cbox_screen, lbl_screen, RIGHT, 1, 1)
        grid_position.attach_next_to(lbl_positions, lbl_screen, BOTTOM, 1, 1)
        grid_position.attach_next_to(hbox_position, lbl_positions, RIGHT, 1, 1)
        grid_position.attach_next_to(btn_sel_geom, lbl_positions, BOTTOM, 2, 1)

        frm_aspect = Gtk.Frame(label_widget=Gtk.Label("<b>%s</b>" % _("Font"),
                                                      use_markup=True),
                               border_width=4,
                               shadow_type=Gtk.ShadowType.NONE,
                               margin=6, hexpand=True)
        grid_aspect = Gtk.Grid(row_spacing=6, column_spacing=6,
                               margin=6)

        frm_position.add(grid_position)

        lbl_font = Gtk.Label(_("Font"),
                             hexpand=True, halign=START)
        btn_font = Gtk.FontButton(self.options.font_desc,
                                  font=self.options.font_desc,
                                  use_font=True, show_size=False)
        if Gtk.check_version(3, 23, 0) is None:
            btn_font.set_level(Gtk.FontChooserLevel.STYLE)
        btn_font.connect("font-set", on_btn_font)

        lbl_sizes = Gtk.Label(_("Size"),
                              halign=START)
        cbox_sizes = Gtk.ComboBoxText(name='size')
        for id_, text in FONT_SIZES.items():
            cbox_sizes.append(id_, text)
            if id_ == self.options.font_size:
                cbox_sizes.props.active_id = id_
        cbox_sizes.connect("changed", on_cbox_sizes_changed)

        grid_aspect.add(lbl_font)
        grid_aspect.attach_next_to(btn_font, lbl_font, RIGHT, 1, 1)
        grid_aspect.attach_next_to(lbl_sizes, lbl_font, BOTTOM, 1, 1)
        grid_aspect.attach_next_to(cbox_sizes, lbl_sizes, RIGHT, 1, 1)
        frm_aspect.add(grid_aspect)

        frm_kbd = Gtk.Frame(label_widget=Gtk.Label("<b>%s</b>" % _("Keys"),
                                                   use_markup=True),
                            border_width=4,
                            shadow_type=Gtk.ShadowType.NONE,
                            margin=6)
        grid_kbd = Gtk.Grid(row_spacing=6, column_spacing=6,
                            margin=6)

        lbl_modes = Gtk.Label(_("Keyboard mode"),
                              halign=START)
        cbox_modes = Gtk.ComboBoxText(name='mode')
        for id_, text in KEY_MODES.items():
            cbox_modes.append(id_, text)
            if id_ == self.options.key_mode:
                cbox_modes.props.active_id = id_
        cbox_modes.connect("changed", on_cbox_modes_changed)

        lbl_bak = Gtk.Label(_("Backspace mode"),
                            halign=START)
        cbox_bak = Gtk.ComboBoxText()
        for id_, text in BAK_MODES.items():
            cbox_bak.append(id_, text)
            if id_ == self.options.bak_mode:
                cbox_bak.props.active_id = id_
        cbox_bak.connect("changed", on_cbox_bak_changed)

        lbl_mods = Gtk.Label(_("Modifiers mode"),
                             halign=START)
        cbox_mods = Gtk.ComboBoxText()
        for id_, text in MODS_MODES.items():
            cbox_mods.append(id_, text)
            if id_ == self.options.mods_mode:
                cbox_mods.props.active_id = id_
        cbox_mods.connect("changed", on_cbox_mods_changed)

        chk_modsonly = Gtk.CheckButton(_("Show Modifier sequences only"),
                                       active=self.options.mods_only)
        chk_modsonly.connect("toggled", on_cbox_modsonly_changed)

        chk_visshift = Gtk.CheckButton(_("Always show Shift"),
                                       active=self.options.vis_shift)
        chk_visshift.connect("toggled", on_cbox_visshift_changed)

        chk_visspace = Gtk.CheckButton(_("Show Whitespace characters"),
                                       active=self.options.vis_space)
        chk_visspace.connect("toggled", on_cbox_visspace_changed)

        hbox_compr = Gtk.Grid(column_spacing=6)
        chk_compr = Gtk.CheckButton(_("Compress repeats after"),
                                    active=self.options.compr_cnt > 0)
        chk_compr.connect("toggled", on_cbox_compr_changed)
        self.sb_compr = Gtk.SpinButton(digits=0,
                                       numeric=True,
                                       update_policy=IF_VALID)
        self.sb_compr.set_increments(1, 1)
        self.sb_compr.set_range(1, 100)
        self.sb_compr.set_value(self.options.compr_cnt or 3)
        self.sb_compr.connect("value-changed", on_sb_compr_changed)
        hbox_compr.add(chk_compr)
        hbox_compr.add(self.sb_compr)

        grid_kbd.add(lbl_modes)
        grid_kbd.attach_next_to(cbox_modes, lbl_modes, RIGHT, 1, 1)
        grid_kbd.attach_next_to(lbl_bak, lbl_modes, BOTTOM, 1, 1)
        grid_kbd.attach_next_to(cbox_bak, lbl_bak, RIGHT, 1, 1)
        grid_kbd.attach_next_to(lbl_mods, lbl_bak, BOTTOM, 1, 1)
        grid_kbd.attach_next_to(cbox_mods, lbl_mods, RIGHT, 1, 1)
        grid_kbd.attach_next_to(chk_modsonly, lbl_mods, BOTTOM, 2, 1)
        grid_kbd.attach_next_to(chk_visshift, chk_modsonly, BOTTOM, 2, 1)
        grid_kbd.attach_next_to(chk_visspace, chk_visshift, BOTTOM, 2, 1)
        grid_kbd.attach_next_to(hbox_compr, chk_visspace, BOTTOM, 2, 1)
        frm_kbd.add(grid_kbd)

        frm_color = Gtk.Frame(label_widget=Gtk.Label("<b>%s</b>" % _("Color"),
                                                     use_markup=True),
                              border_width=4,
                              shadow_type=Gtk.ShadowType.NONE,
                              margin=6)
        grid_color = Gtk.Grid(orientation=VERTICAL,
                              row_spacing=6, column_spacing=6,
                              margin=6)

        lbl_font_color = Gtk.Label(_("Font color"),
                                   halign=START)
        btn_font_color = Gtk.ColorButton(color=Gdk.color_parse(self.options.font_color),
                                         title=_("Text color"),
                                         halign=END)
        btn_font_color.connect("color-set", on_font_color_changed)

        lbl_bg_color = Gtk.Label(_("Background color"),
                                 halign=START)
        btn_bg_color = Gtk.ColorButton(color=Gdk.color_parse(self.options.bg_color),
                                       title=_("Background color"),
                                       halign=END)
        btn_bg_color.connect("color-set", on_bg_color_changed)

        lbl_opacity = Gtk.Label(_("Opacity"),
                                halign=START)
        adj_opacity = Gtk.Adjustment(self.options.opacity, 0, 1.0, 0.1, 0, 0)
        adj_opacity.connect("value-changed", on_adj_opacity_changed)
        adj_scale = Gtk.Scale(adjustment=adj_opacity,
                              hexpand=True, halign=FILL)

        grid_color.add(lbl_font_color)
        grid_color.attach_next_to(btn_font_color, lbl_font_color, RIGHT, 1, 1)
        grid_color.attach_next_to(lbl_bg_color, lbl_font_color, BOTTOM, 1, 1)
        grid_color.attach_next_to(btn_bg_color, lbl_bg_color, RIGHT, 1, 1)
        grid_color.attach_next_to(lbl_opacity, lbl_bg_color, BOTTOM, 1, 1)
        grid_color.attach_next_to(adj_scale, lbl_opacity, RIGHT, 1, 1)
        frm_color.add(grid_color)

        frm_mouse = Gtk.Frame(label_widget=Gtk.Label("<b>%s</b>" % _("Mouse"),
                                                    use_markup=True),
                             border_width=4,
                             shadow_type=Gtk.ShadowType.NONE,
                             margin=6, hexpand=True)
        vbox_mouse = Gtk.VBox(spacing=6)

        chk_mouse = Gtk.CheckButton(_("Show Mouse"))
        chk_mouse.connect("toggled", on_cbox_mouse_changed)
        chk_mouse.set_active(self.options.mouse)
        vbox_mouse.pack_start(chk_mouse, expand=False, fill=True, padding=0)

        hbox_mouse = Gtk.HBox()
        lbl_mouse1 = Gtk.Label(_("Hide duration"))
        lbl_mouse2 = Gtk.Label(_("seconds"))
        sb_mouse = Gtk.SpinButton(digits=1)
        sb_mouse.set_increments(0.5, 1.0)
        sb_mouse.set_range(0.0, 2.0)
        sb_mouse.set_numeric(True)
        sb_mouse.set_update_policy(Gtk.SpinButtonUpdatePolicy.IF_VALID)
        sb_mouse.set_value(self.options.button_hide_duration)
        sb_mouse.connect("value-changed", on_sb_mouse_duration_changed)
        hbox_mouse.pack_start(lbl_mouse1, expand=False, fill=False, padding=6)
        hbox_mouse.pack_start(sb_mouse, expand=False, fill=False, padding=4)
        hbox_mouse.pack_start(lbl_mouse2, expand=False, fill=False, padding=4)
        vbox_mouse.pack_start(hbox_mouse, expand=False, fill=False, padding=6)

        frm_mouse.add(vbox_mouse)
        frm_mouse.show_all()

        hbox_main = Gtk.Grid(column_homogeneous=True)
        vbox_main = Gtk.Grid(orientation=VERTICAL)
        vbox_main.add(frm_time)
        vbox_main.add(frm_position)
        vbox_main.add(frm_aspect)
        hbox_main.add(vbox_main)
        vbox_main = Gtk.Grid(orientation=VERTICAL)
        vbox_main.add(frm_kbd)
        vbox_main.add(frm_color)
        vbox_main.add(frm_mouse)
        hbox_main.add(vbox_main)

        box = prefs.get_content_area()
        box.add(hbox_main)
        box.show_all()


    def make_menu(self):
        self.menu = menu = Gtk.Menu()

        show_item = Gtk.CheckMenuItem(_("Show keys"))
        show_item.set_active(True)
        show_item.connect("toggled", self.on_show_keys)
        show_item.show()
        menu.append(show_item)

        preferences_item = Gtk.MenuItem(_("Preferences"))
        preferences_item.connect("activate", self.on_preferences_dialog)
        preferences_item.show()
        menu.append(preferences_item)

        about_item = Gtk.MenuItem(_("About"))
        about_item.connect("activate", self.on_about_dialog)
        about_item.show()
        menu.append(about_item)

        separator_item = Gtk.SeparatorMenuItem()
        separator_item.show()
        menu.append(separator_item)

        image = Gtk.MenuItem(_("Quit"))
        image.connect("activate", self.quit)
        image.show()
        menu.append(image)
        menu.show()


    def make_appindicator(self):
        from gi.repository import AppIndicator3 as AppIndicator
        self.systray = AppIndicator.Indicator.new(
            APP_NAME, 'indicator-messages', AppIndicator.IndicatorCategory.APPLICATION_STATUS)
        self.systray.set_status(AppIndicator.IndicatorStatus.ACTIVE)
        self.systray.set_attention_icon("indicator-messages-new")
        self.systray.set_icon("preferences-desktop-keyboard-shortcuts")
        self.systray.set_menu(self.menu)
        self.logger.debug("Using AppIndicator.")

    def make_systray(self):
        self.systray = Gtk.StatusIcon()
        self.systray.set_from_icon_name("preferences-desktop-keyboard-shortcuts")
        self.systray.connect("popup-menu", self.on_statusicon_popup, self.menu)
        self.logger.debug("Using StatusIcon.")


    def make_about_dialog(self):
        self.about = about = Gtk.AboutDialog()
        about.set_program_name(APP_NAME)
        about.set_version(VERSION)
        about.set_copyright("""
        Copyright(c) 2010-2012: Pablo Seminario <pabluk@gmail.com>
        Copyright(c) 2015-2020: wave++ "Yuri D'Elia" <wavexx@thregr.org>
        Copyright(c) 2019-2020: Yuto Tokunaga <yuntan.sub1@gmail.com>
        """)
        about.set_comments(APP_DESC)
        about.set_documenters(
            ["José María Quiroga <pepelandia@gmail.com>"]
        )
        about.set_website(APP_URL)
        about.set_icon_name('preferences-desktop-keyboard-shortcuts')
        about.set_logo_icon_name('preferences-desktop-keyboard-shortcuts')
        about.connect("response", lambda *_: about.hide_on_delete())
        about.connect("delete-event", lambda *_: about.hide_on_delete())


    def on_about_dialog(self, widget, data=None):
        self.about.show()


    def start_lockscreen_detection(self):
        from re import match
        from threading import Thread
        from dbus import SessionBus
        from dbus.mainloop.glib import DBusGMainLoop

        def filter_bus_message(bus, message):
            message_member = message.get_member()

            if not self.enabled or message_member != "ActiveChanged":
                return

            args_list = message.get_args_list()
            if args_list[0]:
                self.labelmngr.stop()
                self.logger.debug("Lock Screen; Screenkey disabled.")
            else:
                self.restart_labelmanager()
                self.logger.debug("Unlock Screen; Screenkey enabled.")

        def lockscreen_detection_loop():
            DBusGMainLoop(set_as_default=True)
            session_bus = SessionBus()

            signal_interface = None

            for dbus_string in session_bus.list_names():
                bus_name = str(dbus_string)
                if match(r"org\.(\w+)\.ScreenSaver", bus_name):
                    signal_interface = bus_name
                    self.logger.debug(f"DBUS signal interface found: \"{signal_interface}\" ; password should not show when unlocking the screen.")
                    break

            if not signal_interface:
                self.logger.debug("ScreenSaver DBUS signal interface not found; beware: password may show when unlocking the screen!")
                del(session_bus)
                DBusGMainLoop(set_as_default=False)
                return

            session_bus.add_match_string(f"type='signal',interface='{signal_interface}'")
            session_bus.add_message_filter(filter_bus_message)
            mainloop = GLib.MainLoop()
            mainloop.run()

        thread = Thread(target=lockscreen_detection_loop)
        thread.daemon = True
        thread.start()


    def run(self):
        self.start_lockscreen_detection()
        Gtk.main()
        return self.exit_status



# "screenkey" is distributed under GNU GPLv3+, WITHOUT ANY WARRANTY.
# Copyright(c) 2010-2012: Pablo Seminario <pabluk@gmail.com>
# Copyright(c) 2015-2016: wave++ "Yuri D'Elia" <wavexx@thregr.org>
# Copyright(c) 2019-2020: Yuto Tokunaga <yuntan.sub1@gmail.com>

from . import inputlistener
from .inputlistener import InputListener, InputType

from gi.repository import GLib

from collections import namedtuple
from datetime import datetime

# Key replacement data:
#
# bk_stop: stops backspace processing in baked mode, but not full mode
#          these keys generally move the caret, and are also padded with a thin space
# silent:  always stops backspace processing (baked/full mode)
#          these keys generally do not emit output in the text and cannot be processed
# spaced:  strong spacing is required around the symbol

ReplData = namedtuple('ReplData', ['value', 'font', 'suffix'])
KeyRepl  = namedtuple('KeyRepl',  ['bk_stop', 'silent', 'spaced', 'repl'])
KeyData  = namedtuple('KeyData',  ['stamp', 'is_ctrl', 'bk_stop', 'silent', 'spaced', 'markup'])
ButtonData = namedtuple('ButtonData',  ['stamp', 'btn', 'pressed'])

REPLACE_SYMS = {
    # Regular keys
    'Escape':       KeyRepl(True,  True,  True,  _('Esc')),
    'Tab':          KeyRepl(True,  False, False, _('↹')),
    'ISO_Left_Tab': KeyRepl(True,  False, False, _('↹')),
    'Return':       KeyRepl(True,  False, False, _('⏎')),
    'space':        KeyRepl(False, False, False, _('␣')),
    'BackSpace':    KeyRepl(True,  True,  False, _('⌫')),
    'Caps_Lock':    KeyRepl(True,  True,  True,  _('Caps')),
    'F1':           KeyRepl(True,  True,  True,  _('F1')),
    'F2':           KeyRepl(True,  True,  True,  _('F2')),
    'F3':           KeyRepl(True,  True,  True,  _('F3')),
    'F4':           KeyRepl(True,  True,  True,  _('F4')),
    'F5':           KeyRepl(True,  True,  True,  _('F5')),
    'F6':           KeyRepl(True,  True,  True,  _('F6')),
    'F7':           KeyRepl(True,  True,  True,  _('F7')),
    'F8':           KeyRepl(True,  True,  True,  _('F8')),
    'F9':           KeyRepl(True,  True,  True,  _('F9')),
    'F10':          KeyRepl(True,  True,  True,  _('F10')),
    'F11':          KeyRepl(True,  True,  True,  _('F11')),
    'F12':          KeyRepl(True,  True,  True,  _('F12')),
    'Up':           KeyRepl(True,  True,  False, _('↑')),
    'Left':         KeyRepl(True,  True,  False, _('←')),
    'Right':        KeyRepl(True,  True,  False, _('→')),
    'Down':         KeyRepl(True,  True,  False, _('↓')),
    'Prior':        KeyRepl(True,  True,  True,  _('PgUp')),
    'Next':         KeyRepl(True,  True,  True,  _('PgDn')),
    'Home':         KeyRepl(True,  True,  True,  _('Home')),
    'End':          KeyRepl(True,  True,  True,  _('End')),
    'Insert':       KeyRepl(False, True,  True,  _('Ins')),
    'Delete':       KeyRepl(True,  False, True,  _('Del')),
    'KP_End':       KeyRepl(False, False, True,  _('1ᴷᴾ')),
    'KP_Down':      KeyRepl(False, False, True,  _('2ᴷᴾ')),
    'KP_Next':      KeyRepl(False, False, True,  _('3ᴷᴾ')),
    'KP_Left':      KeyRepl(False, False, True,  _('4ᴷᴾ')),
    'KP_Begin':     KeyRepl(False, False, True,  _('5ᴷᴾ')),
    'KP_Right':     KeyRepl(False, False, True,  _('6ᴷᴾ')),
    'KP_Home':      KeyRepl(False, False, True,  _('7ᴷᴾ')),
    'KP_Up':        KeyRepl(False, False, True,  _('8ᴷᴾ')),
    'KP_Prior':     KeyRepl(False, False, True,  _('9ᴷᴾ')),
    'KP_Insert':    KeyRepl(False, False, True,  _('0ᴷᴾ')),
    'KP_Delete':    KeyRepl(False, False, True,  _('(.)')),
    'KP_Add':       KeyRepl(False, False, True,  _('(+)')),
    'KP_Subtract':  KeyRepl(False, False, True,  _('(-)')),
    'KP_Multiply':  KeyRepl(False, False, True,  _('(*)')),
    'KP_Divide':    KeyRepl(False, False, True,  _('(/)')),
    'KP_Enter':     KeyRepl(True,  False, False, _('⏎')),
    'KP_1':         KeyRepl(False, False, True,  _('1ᴷᴾ')),
    'KP_2':         KeyRepl(False, False, True,  _('2ᴷᴾ')),
    'KP_3':         KeyRepl(False, False, True,  _('3ᴷᴾ')),
    'KP_4':         KeyRepl(False, False, True,  _('4ᴷᴾ')),
    'KP_5':         KeyRepl(False, False, True,  _('5ᴷᴾ')),
    'KP_6':         KeyRepl(False, False, True,  _('6ᴷᴾ')),
    'KP_7':         KeyRepl(False, False, True,  _('7ᴷᴾ')),
    'KP_8':         KeyRepl(False, False, True,  _('8ᴷᴾ')),
    'KP_9':         KeyRepl(False, False, True,  _('9ᴷᴾ')),
    'KP_0':         KeyRepl(False, False, True,  _('0ᴷᴾ')),
    'Num_Lock':     KeyRepl(False, True,  True,  _('NumLck')),
    'Scroll_Lock':  KeyRepl(False, True,  True,  _('ScrLck')),
    'Pause':        KeyRepl(False, True,  True,  _('Pause')),
    'Break':        KeyRepl(False, True,  True,  _('Break')),
    'Print':        KeyRepl(False, True,  True,  _('Print')),
    'Multi_key':    KeyRepl(False, True,  True,  _('Compose')),

    # Multimedia keys
    'XF86AudioMute':         KeyRepl(True, True, True, [ReplData(_('\uf026'),    'Font Awesome 5 Free', None),
                                                        ReplData(_('\uf026'),    'FontAwesome',         None),
                                                        ReplData(_('Mute'),      None,                  None)]),
    'XF86AudioMicMute':      KeyRepl(True, True, True, [ReplData(_('\uf131'),    'Font Awesome 5 Free', None),
                                                        ReplData(_('\uf131'),    'FontAwesome',         None),
                                                        ReplData(_('Rec'),       None,                  None)]),
    'XF86AudioRaiseVolume':  KeyRepl(True, True, True, [ReplData(_('\uf028'),    'Font Awesome 5 Free', None),
                                                        ReplData(_('\uf028'),    'FontAwesome',         None),
                                                        ReplData(_('Vol'),       None,                  '+')]),
    'XF86AudioLowerVolume':  KeyRepl(True, True, True, [ReplData(_('\uf027'),    'Font Awesome 5 Free', None),
                                                        ReplData(_('\uf027'),    'FontAwesome',         None),
                                                        ReplData(_('Vol'),       None,                  '-')]),
    'XF86AudioPrev':         KeyRepl(True, True, True, [ReplData(_('\uf048'),    'Font Awesome 5 Free', None),
                                                        ReplData(_('\uf048'),    'FontAwesome',         None),
                                                        ReplData(_('Prev'),      None,                  None)]),
    'XF86AudioNext':         KeyRepl(True, True, True, [ReplData(_('\uf051'),    'Font Awesome 5 Free', None),
                                                        ReplData(_('\uf051'),    'FontAwesome',         None),
                                                        ReplData(_('Next'),      None,                  None)]),
    'XF86AudioPlay':         KeyRepl(True, True, True, [ReplData(_('\uf04b'),    'Font Awesome 5 Free', None),
                                                        ReplData(_('\uf04b'),    'FontAwesome',         None),
                                                        ReplData(_('▶'),         None,                  None)]),
    'XF86AudioStop':         KeyRepl(True, True, True, [ReplData(_('\uf04d'),    'Font Awesome 5 Free', None),
                                                        ReplData(_('\uf04d'),    'FontAwesome',         None),
                                                        ReplData(_('⬛'),         None,                  None)]),
    'XF86Eject':             KeyRepl(True, True, True, [ReplData(_('\uf052'),    'Font Awesome 5 Free', None),
                                                        ReplData(_('\uf052'),    'FontAwesome',         None),
                                                        ReplData(_('Eject'),     None,                  None)]),
    'XF86MonBrightnessDown': KeyRepl(True, True, True, [ReplData(_('\uf185'),    'Font Awesome 5 Free', '-'),
                                                        ReplData(_('\uf185'),    'FontAwesome',         '-'),
                                                        ReplData(_('Bright'),    None,                  '-')]),
    'XF86MonBrightnessUp':   KeyRepl(True, True, True, [ReplData(_('\uf185'),    'Font Awesome 5 Free', '+'),
                                                        ReplData(_('\uf185'),    'FontAwesome',         '+'),
                                                        ReplData(_('Bright'),    None,                  '+')]),
    'XF86Display':           KeyRepl(True, True, True, [ReplData(_('\uf108'),    'Font Awesome 5 Free', None),
                                                        ReplData(_('\uf108'),    'FontAwesome',         None),
                                                        ReplData(_('Display'),   None,                  None)]),
    'XF86WLAN':              KeyRepl(True, True, True, [ReplData(_('\uf1eb'),    'Font Awesome 5 Free', None),
                                                        ReplData(_('\uf1eb'),    'FontAwesome',         None),
                                                        ReplData(_('WLAN'),      None,                  None)]),
    'XF86Search':            KeyRepl(True, True, True, [ReplData(_('\uf002'),    'Font Awesome 5 Free', None),
                                                        ReplData(_('\uf002'),    'FontAwesome',         None),
                                                        ReplData(_('Search'),    None,                  None)]),
    'XF86Bluetooth':         KeyRepl(True, True, True, [ReplData(_('\uf294'),    'Font Awesome 5 Free', None),
                                                        ReplData(_('\uf294'),    'FontAwesome',         None),
                                                        ReplData(_('Bluetooth'), None,                  None)]),
    'XF86Tools':             KeyRepl(True, True, True, [ReplData(_('\uf7d9'),    'Font Awesome 5 Free', None),
                                                        ReplData(_('🛠'),        None,                  None)]),
    'XF86Favorites':         KeyRepl(True, True, True, [ReplData(_('\uf005'),    'Font Awesome 5 Free', None),
                                                        ReplData(_('\uf005'),    'FontAwesome',         None),
                                                        ReplData(_('🟊'),        None,                  None)]),
}

WHITESPACE_SYMS = {'Tab', 'ISO_Left_Tab', 'Return', 'space', 'KP_Enter'}

MODS_SYMS = {
    'shift':  {'Shift_L', 'Shift_R'},
    'ctrl':   {'Control_L', 'Control_R'},
    'alt':    {'Alt_L', 'Alt_R', 'Meta_L', 'Meta_R'},
    'super':  {'Super_L', 'Super_R'},
    'hyper':  {'Hyper_L', 'Hyper_R'},
    'alt_gr': {'ISO_Level3_Shift'},
}

REPLACE_MODS = {
    'shift':  {'normal': _('Shift+'), 'emacs': 'S-', 'mac': _('⇧+')},
    'ctrl':   {'normal': _('Ctrl+'),  'emacs': 'C-', 'mac': _('⌘+')},
    'alt':    {'normal': _('Alt+'),   'emacs': 'M-', 'mac': _('⌥+')},
    'super':  {'normal': _('Super+'), 'emacs': 's-',
               'win': [ReplData(_('\uf17a'), 'Font Awesome 5 Free', '+'),
                       ReplData(_('\uf17a'), 'FontAwesome',         '+'),
                       ReplData(_('Win'),    None,                  '+')],
               'tux': [ReplData(_('\uf17c'), 'Font Awesome 5 Free', '+'),
                       ReplData(_('\uf17c'), 'FontAwesome',         '+'),
                       ReplData(_('Super'),  None,                  '+')]},
    'hyper':  {'normal': _('Hyper+'), 'emacs': 'H-'},
    'alt_gr': {'normal': _('AltGr+'), 'emacs': 'AltGr-'},
}


def keysym_to_mod(keysym):
    for k, v in MODS_SYMS.items():
        if keysym in v:
            return k
    return None


class LabelManager:
    def __init__(self, label_listener, image_listener, logger, key_mode,
                 bak_mode, mods_mode, mods_only, multiline, vis_shift,
                 vis_space, recent_thr, compr_cnt, ignore, pango_ctx,
                 enabled):
        self.key_mode = key_mode
        self.bak_mode = bak_mode
        self.mods_mode = mods_mode
        self.logger = logger
        self.label_listener = label_listener
        self.image_listener = image_listener
        self.data = []
        self.enabled = enabled
        self.mods_only = mods_only
        self.multiline = multiline
        self.vis_shift = vis_shift
        self.vis_space = vis_space
        self.recent_thr = recent_thr
        self.compr_cnt = compr_cnt
        self.ignore = ignore
        self.kl = None
        self.font_families = {x.get_name() for x in pango_ctx.list_families()}
        self.update_replacement_map()


    def __del__(self):
        self.stop()


    def start(self):
        self.stop()
        compose = (self.key_mode == 'composed')
        translate = (self.key_mode in ['composed', 'translated'])
        self.kl = InputListener(self.event_handler,
                                InputType.keyboard | InputType.button,
                                compose, translate)
        self.kl.start()
        self.logger.debug("Thread started.")


    def stop(self):
        if self.kl:
            self.kl.stop()
            self.logger.debug("Thread stopped.")
            self.kl.join()
            self.kl = None


    def clear(self):
        self.data = []


    def get_repl_markup(self, repl):
        if type(repl) != list:
            repl = [repl]
        for c in repl:
            # no replacement data
            if type(c) != ReplData:
                return GLib.markup_escape_text(c)

            # plain suffix
            if c.suffix is None:
                sfx = ''
            else:
                sfx = GLib.markup_escape_text(c.suffix)

            if c.font is None:
                # regular font
                return GLib.markup_escape_text(c.value) + sfx;
            elif c.font in self.font_families:
                # custom symbol
                return '<span font_family="' + c.font + '" font_weight="regular">' + \
                    GLib.markup_escape_text(c.value) + '</span>' + sfx;


    def update_replacement_map(self):
        self.replace_syms = {}
        for k, v in REPLACE_SYMS.items():
            markup = self.get_repl_markup(v.repl)
            self.replace_syms[k] = KeyRepl(v.bk_stop, v.silent, v.spaced, markup)

        self.replace_mods = {}
        for k, v in REPLACE_MODS.items():
            data = v.get(self.mods_mode, v['normal'])
            self.replace_mods[k] = self.get_repl_markup(data)


    def update_text(self, synthetic=False):
        markup = ""
        recent = False
        stamp = datetime.now()
        repeats = 0
        for i, key in enumerate(self.data):
            if i != 0:
                last = self.data[i - 1]

                # compress repeats
                if self.compr_cnt and key.markup == last.markup:
                    repeats += 1
                    if repeats < self.compr_cnt:
                        pass
                    elif i == len(self.data) - 1 or key.markup != self.data[i + 1].markup:
                        if not recent and (stamp - key.stamp).total_seconds() < self.recent_thr:
                            markup += '<u>'
                            recent = True
                        markup += '<sub><small>…{}×</small></sub>'.format(repeats + 1)
                        if len(key.markup) and key.markup[-1] == '\n':
                            markup += '\n'
                        continue
                    else:
                        continue

                # character block spacing
                if len(last.markup) and last.markup[-1] == '\n':
                    pass
                elif key.is_ctrl or last.is_ctrl or key.spaced or last.spaced:
                    markup += ' '
                elif key.bk_stop or last.bk_stop or repeats > self.compr_cnt:
                    markup += '<span font_family="sans">\u2009</span>'
                if key.markup != last.markup:
                    repeats = 0

            key_markup = key.markup
            if type(key_markup) is bytes:
                key_markup = key_markup.decode()
            if not recent and (stamp - key.stamp).total_seconds() < self.recent_thr:
                recent = True
                key_markup = '<u>' + key_markup

            # disable ligatures
            if len(key.markup) == 1 and 0x0300 <= ord(key.markup) <= 0x036F:
                # workaround for pango not handling ZWNJ correctly for combining marks
                markup += '\u180e' + key_markup + '\u200a'
            elif len(key_markup):
                markup += '\u200c' + key_markup

        if len(markup) and markup[-1] == '\n':
            markup = markup.rstrip('\n')
            if not self.vis_space and not self.data[-1].is_ctrl:
                # always show some return symbol at the last line
                markup += self.replace_syms['Return'].repl
        if recent:
            markup += '</u>'
        self.logger.debug("Label updated: %s." % repr(markup))
        self.label_listener(markup, synthetic)


    def queue_update(self):
        self.update_text(True)


    def event_handler(self, event):
        if event is None:
            self.logger.debug("inputlistener failure: {}".format(str(self.kl.error)))
            self.label_listener(None, None)
            return

        if isinstance(event, inputlistener.KeyData):
            self.key_press(event)
        elif isinstance(event, inputlistener.ButtonData):
            self.btn_press(event)
        else:
            self.logger.error("unhandled event type {}".format(type(event)))


    def key_press(self, event):
        if event.symbol is None:
            # TODO: Investigate what causes this to happen.
            # I caught it once in pdb, but in this function, not in inputlistener,
            # and KeyData doesn't contain enough info.
            return
        symbol = event.symbol.decode()

        if self.enabled:
            for mod, button_id in zip(['ctrl', 'alt', 'shift'], range(8, 11)):
                if symbol in MODS_SYMS[mod]:
                    self.image_listener(ButtonData(
                        datetime.now(), button_id, event.pressed
                    ))

        if event.pressed == False:
            self.logger.debug("Key released {:5}(ks): {}".format(event.keysym, symbol))
            return
        if symbol in self.ignore:
            self.logger.debug("Key ignored  {:5}(ks): {}".format(event.keysym, symbol))
            return
        if event.filtered:
            self.logger.debug("Key filtered {:5}(ks): {}".format(event.keysym, symbol))
        else:
            state = "repeated" if event.repeated else "pressed"
            string = repr(event.string)
            self.logger.debug("Key {:8} {:5}(ks): {} ({}, mask: {:08b})".format
                              (state, event.keysym, string, symbol, event.mods_mask))

        # Stealth enable/disable handling
        for mod in ['shift', 'ctrl', 'alt']:
            if not event.repeated and event.modifiers[mod] \
               and symbol in MODS_SYMS[mod]:
                self.enabled = not self.enabled
                state = 'enabled' if self.enabled else 'disabled'
                if not self.enabled:
                    self.image_listener(None)
                self.logger.info("{mod}+{mod} detected: screenkey {state}".format(
                    mod=mod.capitalize(), state=state))
        if not self.enabled:
            return False

        # keep the window alive as the user is composing
        mod_pressed = keysym_to_mod(symbol) is not None
        update = len(self.data) and (event.filtered or mod_pressed)

        if not event.filtered:
            if self.key_mode in ['translated', 'composed']:
                update |= self.key_normal_mode(event)
            elif self.key_mode == 'raw':
                update |= self.key_raw_mode(event)
            else:
                update |= self.key_keysyms_mode(event)
        if update:
            self.update_text()


    def key_normal_mode(self, event):
        self.logger.debug("key_normal_mode")
        # Visible modifiers
        mod = ''
        for cap in ['ctrl', 'alt', 'super', 'hyper']:
            if event.modifiers[cap]:
                mod = mod + self.replace_mods[cap]

        # Backspace handling
        symbol = event.symbol.decode()
        if symbol == 'BackSpace' and not self.mods_only and \
           mod == '' and not event.modifiers['shift']:
            key_repl = self.replace_syms.get(symbol)
            if self.bak_mode == 'normal':
                self.data.append(KeyData(datetime.now(), False, *key_repl))
                return True
            else:
                if not len(self.data):
                    pop = False
                else:
                    last = self.data[-1]
                    if last.is_ctrl:
                        pop = False
                    elif self.bak_mode == 'baked':
                        pop = not last.bk_stop
                    else:
                        pop = not last.silent
                if pop:
                    self.data.pop()
                else:
                    self.data.append(KeyData(datetime.now(), False, *key_repl))
                return True

        # Regular keys
        key_repl = self.replace_syms.get(symbol)
        replaced = key_repl is not None
        if key_repl is None:
            if keysym_to_mod(symbol):
                return False
            else:
                repl = event.string or symbol
                markup = GLib.markup_escape_text(repl)
                key_repl = KeyRepl(False, False, len(repl) > 1, markup)

        if event.modifiers['shift'] and \
           (replaced or (mod != '' and \
                         self.vis_shift and \
                         self.mods_mode != 'emacs')):
            # add back shift for translated keys
            mod = mod + self.replace_mods['shift']

        # Whitespace handling
        if not self.vis_space and mod == '' and symbol in WHITESPACE_SYMS:
            if symbol not in ['Return', 'KP_Enter']:
                repl = event.string
            elif self.multiline:
                repl = ''
            else:
                repl = key_repl.repl
            key_repl = KeyRepl(key_repl.bk_stop, key_repl.silent, key_repl.spaced, repl)

        # Multiline
        if symbol in ['Return', 'KP_Enter'] and self.multiline == True:
            key_repl = KeyRepl(key_repl.bk_stop, key_repl.silent,
                               key_repl.spaced, key_repl.repl + '\n')

        if mod == '':
            if not self.mods_only:
                repl = key_repl.repl

                # switches
                if symbol in ['Caps_Lock', 'Num_Lock']:
                    state = event.modifiers[symbol.lower()]
                    repl += '(%s)' % (_('off') if state else _('on'))

                self.data.append(KeyData(datetime.now(), False, key_repl.bk_stop,
                                         key_repl.silent, key_repl.spaced, repl))
                return True
        else:
            if self.mods_mode == 'emacs' or key_repl.repl[0] != mod[-1]:
                repl = mod + key_repl.repl
            else:
                repl = mod + '‟' + key_repl.repl + '”'
            self.data.append(KeyData(datetime.now(), True, key_repl.bk_stop,
                                     key_repl.silent, key_repl.spaced, repl))
            return True

        return False


    def key_raw_mode(self, event):
        # modifiers
        mod = ''
        for cap in REPLACE_MODS.keys():
            if event.modifiers[cap]:
                mod = mod + self.replace_mods[cap]

        # keycaps
        symbol = event.symbol.decode()
        key_repl = self.replace_syms.get(symbol)
        if key_repl is None:
            if keysym_to_mod(symbol):
                return False
            else:
                repl = event.string.upper() if event.string else symbol
                markup = GLib.markup_escape_text(repl)
                key_repl = KeyRepl(False, False, len(repl) > 1, markup)

        if mod == '':
            repl = key_repl.repl

            # switches
            if symbol in ['Caps_Lock', 'Num_Lock']:
                state = event.modifiers[symbol.lower()]
                repl += '(%s)' % (_('off') if state else _('on'))

            self.data.append(KeyData(datetime.now(), False, key_repl.bk_stop,
                                     key_repl.silent, key_repl.spaced, repl))
        else:
            if self.mods_mode == 'emacs' or key_repl.repl[0] != mod[-1]:
                repl = mod + key_repl.repl
            else:
                repl = mod + '‟' + key_repl.repl + '”'
            self.data.append(KeyData(datetime.now(), True, key_repl.bk_stop,
                                     key_repl.silent, key_repl.spaced, repl))
        return True


    def key_keysyms_mode(self, event):
        symbol = event.symbol.decode()
        if symbol in REPLACE_SYMS:
            value = symbol
        else:
            value = event.string or symbol
        self.data.append(KeyData(datetime.now(), True, True, True, True, value))
        return True

    # callback for mouse button presses
    def btn_press(self, event):
        if not self.enabled:
            return False
        if event.pressed:
            action = "pressed"
        else:
            action = "released"
        self.logger.debug("Mouse button %d %s" % (event.btn, action))

        # possible event.btn values:
        # 1 = LMB, 2 = MMB (wheel click), 3 = RMB, 4/5 = wheel up/down,
        # 6/7 = wheel left/right, 8+ extra buttons (e. g. thumb buttons)
        if event.btn > 7:
            # image_listener can only handle common buttons, redirect thumb buttons etc. to label
            if not event.pressed:
                return False
            # what we refer to as "Mouse 4" has an internal value of 8,
            # so we subtract 4 from the btn value
            markup = GLib.markup_escape_text("M{}".format(event.btn - 4))

            # show as label, treated the same as keyboard button presses
            self.data.append(KeyData(datetime.now(), False, True,
                         True, True, markup))
            self.update_text()
        else:
            # show event in image
            self.image_listener(
                ButtonData(datetime.now(), event.btn, event.pressed)
            )



# Distributed under the GNU GPLv3+ license, WITHOUT ANY WARRANTY.
# Copyright(c) 2015: wave++ "Yuri D'Elia" <wavexx@thregr.org>
#
# Outputting translated X11 keystrokes is not a simple problem as soon as XIM
# is introduced: getting an updated keyboard/modifier map is not enough to
# replicate the [complex] logic hidden in the input method.
#
# For this reason we use a fairly convoluted mechanism: we record keystrokes
# using the XRecord extension, but we relay them to another fake window running
# on the same server. By manipulating the event, we trick the input method to
# perform composition for us, and poll for translated output events using
# Xutf8LookupString. Since we cannot determine the state of the input context
# for the target window (we're recording blindly), we also need to reset
# context state carefully when the user switches the active focus. 3(!) extra
# connections to the display server are required for this task, and since we're
# using blocking APIs, having to run on our own thread means we cannot share
# any of those with the regular process. On the other hand, many other keycode
# translation issues are avoided by using the string lookup directly.
#
# This is, of course, never going to be always identical to the final output,
# since we're guessing the state of the client (we do the same when guessing
# the result of backspace anyway). But incidentally this method would also
# allow us to poll the input mechanism while composing, to better reflect the
# actual typing on the keyboard.
#
# Some of the code /could/ have been simplified by using XCB for protocol
# translation, but since there's no equivalent to XKB/XIM, I found the exercise
# futile. Needing to use XIM directly also barred pure-python equivalents. As
# a result, we have to drop back to ctypes for _extra_ phun.
#
# Drop me a line if you ever find this comment helpful, as finding a decent
# solution was not trivial -- YD 21/08/2015.

if __name__ == '__main__':
    import xlib
    import keysyms
else:
    from . import xlib
    from . import keysyms

import sys
if sys.version_info.major < 3:
    import glib
else:
    from gi.repository import GLib as glib

import threading
import warnings
import select


# convenience wrappers
def coalesce_ranges(ranges):
    ranges = sorted(ranges, key=lambda x: x[0])
    ret = ranges[:1]
    for r in ranges[1:]:
        if ret[-1][1] < r[0] - 1:
            ret.append(r)
        else:
            ret[-1][1] = max(ret[-1][1], r[1])
    return ret


def record_context(dpy, ev_ranges, dev_ranges):
    ev_ranges = coalesce_ranges(ev_ranges)
    dev_ranges = coalesce_ranges(dev_ranges)

    specs = max(len(ev_ranges), len(dev_ranges))
    range_specs = (xlib.POINTER(xlib.XRecordRange) * specs)()

    for i in range(specs):
        range_specs[i] = xlib.XRecordAllocRange()
        if len(ev_ranges) > i:
            range_specs[i].contents.delivered_events.first = ev_ranges[i][0]
            range_specs[i].contents.delivered_events.last = ev_ranges[i][1]
        if len(dev_ranges) > i:
            range_specs[i].contents.device_events.first = dev_ranges[i][0]
            range_specs[i].contents.device_events.last = dev_ranges[i][1]

    rec_ctx = xlib.XRecordCreateContext(
        dpy, 0,
        xlib.byref(xlib.c_ulong(xlib.XRecordAllClients)), 1,
        range_specs, specs)

    for i in range(specs):
        xlib.XFree(range_specs[i])

    return rec_ctx


def record_enable(dpy, rec_ctx, callback):
    def intercept(data):
        if data.category != xlib.XRecordFromServer:
            return
        if data.client_swapped:
            warnings.warn("cannot handle swapped protocol data")
            return
        ev = xlib.XWireToEvent(dpy, data.data)
        callback(ev)

    def intercept_(_, data):
        intercept(data.contents)
        xlib.XRecordFreeData(data)

    proc = xlib.XRecordInterceptProc(intercept_)
    xlib.XRecordEnableContextAsync(dpy, rec_ctx, proc, None)
    return proc


def create_replay_window(dpy):
    win_attr = xlib.XSetWindowAttributes()
    win_attr.override_redirect = True
    win = xlib.XCreateWindow(dpy, xlib.XDefaultRootWindow(dpy),
                             0, 0, 1, 1, 0,
                             xlib.CopyFromParent, xlib.InputOnly, None,
                             xlib.CWOverrideRedirect,
                             xlib.byref(win_attr))
    return win


def phantom_release(dpy, kev):
    if not xlib.XPending(dpy):
        return False
    ev = xlib.XEvent()
    xlib.XPeekEvent(dpy, xlib.byref(ev))
    return (ev.type == xlib.KeyPress and \
            ev.xkey.state == kev.state and \
            ev.xkey.keycode == kev.keycode and \
            ev.xkey.time == kev.time)


def keysym_to_unicode(keysym):
    if 0x01000000 <= keysym <= 0x0110FFFF:
        return unichr(keysym - 0x01000000)
    keydata = keysyms.KEYSYMS.get(keysym)
    if keydata is not None:
        return keydata[0]
    return None



class KeyData():
    def __init__(self, pressed=None, filtered=None, repeated=None,
                 string=None, keysym=None, status=None, symbol=None,
                 mods_mask=None, modifiers=None):
        self.pressed = pressed
        self.filtered = filtered
        self.repeated = repeated
        self.string = string
        self.keysym = keysym
        self.status = status
        self.symbol = symbol
        self.mods_mask = mods_mask
        self.modifiers = modifiers


class ButtonData():
    def __init__(self, btn, pressed):
        self.btn = btn
        self.pressed = pressed == xlib.ButtonPress


class InputType:
    keyboard = 0b001
    button   = 0b010
    movement = 0b100
    all      = 0b111



class InputListener(threading.Thread):
    def __init__(self, event_callback, input_types=InputType.all,
                 kbd_compose=True, kbd_translate=True):
        super().__init__()
        self.event_callback = event_callback
        self.input_types = input_types
        self.kbd_compose = kbd_compose
        self.kbd_translate = kbd_translate
        self.lock = threading.Lock()
        self.stopped = True
        self.error = None


    def _event_received(self, ev):
        if xlib.KeyPress <= ev.type <= xlib.MotionNotify:
            xlib.XSendEvent(self.replay_dpy, self.replay_win, False, 0, ev)
        elif ev.type in [xlib.FocusIn, xlib.FocusOut]:
            # Forward the event as a custom message in the same queue instead
            # of resetting the XIC directly, in order to preserve queued events
            fwd_ev = xlib.XEvent()
            fwd_ev.type = xlib.ClientMessage
            fwd_ev.xclient.message_type = self.custom_atom
            fwd_ev.xclient.format = 32
            fwd_ev.xclient.data[0] = ev.type
            xlib.XSendEvent(self.replay_dpy, self.replay_win, False, 0, fwd_ev)


    def _event_callback(self, data):
        self.event_callback(data)
        return False

    def _event_processed(self, data):
        data.symbol = xlib.XKeysymToString(data.keysym)
        if data.string is None:
            data.string = keysym_to_unicode(data.keysym)
        glib.idle_add(self._event_callback, data)


    def _event_modifiers(self, kev, data):
        data.modifiers = modifiers = {}
        modifiers['shift'] = bool(kev.state & xlib.ShiftMask)
        modifiers['caps_lock'] = bool(kev.state & xlib.LockMask)
        modifiers['ctrl'] = bool(kev.state & xlib.ControlMask)
        modifiers['alt'] = bool(kev.state & xlib.Mod1Mask)
        modifiers['num_lock'] = bool(kev.state & xlib.Mod2Mask)
        modifiers['hyper'] = bool(kev.state & xlib.Mod3Mask)
        modifiers['super'] = bool(kev.state & xlib.Mod4Mask)
        modifiers['alt_gr'] = bool(kev.state & xlib.Mod5Mask)


    def _event_keypress(self, kev, data):
        buf = xlib.create_string_buffer(16)
        keysym = xlib.KeySym()
        status = xlib.Status()
        ret = xlib.Xutf8LookupString(self._kbd_replay_xic, kev, buf, len(buf),
                                     xlib.byref(keysym), xlib.byref(status))
        if ret != xlib.NoSymbol:
            if 32 <= keysym.value <= 126:
                # avoid ctrl sequences, just take the character value
                data.string = chr(keysym.value)
            else:
                try:
                    data.string = buf.value.decode('utf-8')
                except UnicodeDecodeError:
                    pass
        data.keysym = keysym.value
        data.status = status.value


    def _event_lookup(self, kev, data):
        # this is mostly for debugging: we do not account for group/level
        data.keysym = xlib.XkbKeycodeToKeysym(kev.display, kev.keycode, 0, 0)


    def start(self):
        self.lock.acquire()
        self.stopped = False
        self.error = None
        super().start()


    def stop(self):
        with self.lock:
            if not self.stopped:
                self.stopped = True
                xlib.XRecordDisableContext(self.control_dpy, self.record_ctx)


    def _kbd_init(self):
        self._kbd_last_ev = xlib.XEvent()

        if self.kbd_compose:
            style = xlib.XIMPreeditNothing | xlib.XIMStatusNothing
        else:
            style = xlib.XIMPreeditNone | xlib.XIMStatusNone

        # TODO: implement preedit callbacks for on-the-spot composition
        #       (this would fix focus-stealing for the global IM state)
        self._kbd_replay_xim = xlib.XOpenIM(self.replay_dpy, None, None, None)
        if not self._kbd_replay_xim:
            raise Exception("Cannot initialize input method")

        self._kbd_replay_xic = xlib.XCreateIC(self._kbd_replay_xim,
                                              xlib.XNClientWindow, self.replay_win,
                                              xlib.XNInputStyle, style,
                                              None)
        xlib.XSetICFocus(self._kbd_replay_xic)


    def _kbd_del(self):
        xlib.XDestroyIC(self._kbd_replay_xic)
        xlib.XCloseIM(self._kbd_replay_xim)


    def _kbd_process(self, ev):
        if ev.type == xlib.ClientMessage and \
           ev.xclient.message_type == self.custom_atom:
            if ev.xclient.data[0] in [xlib.FocusIn, xlib.FocusOut]:
                # we do not keep track of multiple XICs, just reset
                xic = xlib.Xutf8ResetIC(self._kbd_replay_xic)
                if xic is not None: xlib.XFree(xic)
            return
        elif ev.type in [xlib.KeyPress, xlib.KeyRelease]:
            # fake keyboard event data for XFilterEvent
            ev.xkey.send_event = False
            ev.xkey.window = self.replay_win

        # pass _all_ events to XFilterEvent
        filtered = bool(xlib.XFilterEvent(ev, 0))
        if ev.type == xlib.KeyRelease and \
           phantom_release(self.replay_dpy, ev.xkey):
            return
        if ev.type not in [xlib.KeyPress, xlib.KeyRelease]:
            return

        # generate new keyboard event
        data = KeyData()
        data.filtered = filtered
        data.pressed = (ev.type == xlib.KeyPress)
        data.repeated = (ev.type == self._kbd_last_ev.type and \
                         ev.xkey.state == self._kbd_last_ev.xkey.state and \
                         ev.xkey.keycode == self._kbd_last_ev.xkey.keycode)
        data.mods_mask = ev.xkey.state
        self._event_modifiers(ev.xkey, data)
        if not data.filtered and data.pressed and self.kbd_translate:
            self._event_keypress(ev.xkey, data)
        else:
            self._event_lookup(ev.xkey, data)
        self._event_processed(data)
        self._kbd_last_ev = ev


    def _btn_process(self, ev):
        if ev.type in [xlib.ButtonPress, xlib.ButtonRelease]:
            data = ButtonData(ev.xbutton.button, ev.type)
            glib.idle_add(self._event_callback, data)


    def run(self):
        # control connection
        self.control_dpy = xlib.XOpenDisplay(None)
        xlib.XSynchronize(self.control_dpy, True)

        # unmapped replay window
        self.replay_dpy = xlib.XOpenDisplay(None)
        self.custom_atom = xlib.XInternAtom(self.replay_dpy, b"SCREENKEY", False)
        replay_fd = xlib.XConnectionNumber(self.replay_dpy)
        self.replay_win = create_replay_window(self.replay_dpy)

        # bail during initialization errors
        try:
            if self.input_types & InputType.keyboard:
                self._kbd_init()
        except Exception as e:
            self.error = e
            xlib.XCloseDisplay(self.control_dpy)
            xlib.XDestroyWindow(self.replay_dpy, self.replay_win)
            xlib.XCloseDisplay(self.replay_dpy)

            # cheap wakeup() equivalent for compatibility
            glib.idle_add(self._event_callback, None)

            self.stopped = True
            self.lock.release()
            return

        # initialize recording context
        ev_ranges = []
        dev_ranges = []
        if self.input_types & InputType.keyboard:
            ev_ranges.append([xlib.FocusIn, xlib.FocusOut])
            dev_ranges.append([xlib.KeyPress, xlib.KeyRelease])
        if self.input_types & InputType.button:
            dev_ranges.append([xlib.ButtonPress, xlib.ButtonRelease])
        if self.input_types & InputType.movement:
            dev_ranges.append([xlib.MotionNotify, xlib.MotionNotify])
        self.record_ctx = record_context(self.control_dpy, ev_ranges, dev_ranges);

        record_dpy = xlib.XOpenDisplay(None)
        record_fd = xlib.XConnectionNumber(record_dpy)
        # we need to keep the record_ref alive(!)
        record_ref = record_enable(record_dpy, self.record_ctx, self._event_received)

        # event loop
        self.lock.release()
        while True:
            with self.lock:
                if self.stopped:
                    break

            r_fd = []
            if xlib.XPending(record_dpy):
                r_fd.append(record_fd)
            if xlib.XPending(self.replay_dpy):
                r_fd.append(replay_fd)
            if not r_fd:
                r_fd, _, _ = select.select([record_fd, replay_fd], [], [])
            if not r_fd:
                break

            if record_fd in r_fd:
                xlib.XRecordProcessReplies(record_dpy)
                xlib.XFlush(self.replay_dpy)

            if replay_fd in r_fd:
                ev = xlib.XEvent()
                xlib.XNextEvent(self.replay_dpy, xlib.byref(ev))
                if self.input_types & InputType.keyboard:
                    self._kbd_process(ev)
                if self.input_types & InputType.button:
                    self._btn_process(ev)

        # finalize
        self.lock.acquire()

        xlib.XRecordFreeContext(self.control_dpy, self.record_ctx)
        xlib.XCloseDisplay(self.control_dpy)
        xlib.XCloseDisplay(record_dpy)
        del record_ref

        if self.input_types & InputType.keyboard:
            self._kbd_del()

        xlib.XDestroyWindow(self.replay_dpy, self.replay_win)
        xlib.XCloseDisplay(self.replay_dpy)

        self.stopped = True
        self.lock.release()



if __name__ == '__main__':
    def callback(data):
        values = {}
        for k in dir(data):
            if k[0] == '_': continue
            values[k] = getattr(data, k)
        print(values)

    kl = InputListener(callback)
    try:
        # keep running only while the listener is alive
        kl.start()
        while kl.is_alive():
            glib.main_context_default().iteration()
    except KeyboardInterrupt:
        pass

    # check if the thread terminated unexpectedly
    if kl.is_alive():
        kl.stop()
        kl.join()
    elif kl.error:
        print("initialization error: {}".format(kl.error))
        if '__traceback__' in dir(kl.error):
            import traceback
            traceback.print_tb(kl.error.__traceback__)
        exit(1)



# Distributed under the GNU GPLv3+ license, WITHOUT ANY WARRANTY.
# Copyright(c) 2015: wave++ "Yuri D'Elia" <wavexx@thregr.org>

from ctypes import *

## base X11
libX11 = CDLL('libX11.so')

# types
Atom = c_ulong
Bool = c_int
XID = c_ulong
Colormap = XID
Cursor = XID
KeyCode = c_ubyte
KeySym = XID
Pixmap = XID
Status = c_int
String = c_char_p
Time = c_ulong
Window = XID

class Display(Structure):
    pass

class Visual(Structure):
    pass

class XKeyEvent(Structure):
    _fields_ = [('type', c_int),
                ('serial', c_ulong),
                ('send_event', Bool),
                ('display', POINTER(Display)),
                ('window', Window),
                ('root', Window),
                ('subwindow', Window),
                ('time', Time),
                ('x', c_int),
                ('y', c_int),
                ('x_root', c_int),
                ('y_root', c_int),
                ('state', c_uint),
                ('keycode', c_uint),
                ('same_screen', Bool)]

XKeyPressedEvent = XKeyEvent
XKeyReleasedEvent = XKeyEvent

class XButtonEvent(Structure):
    _fields_ = [('type', c_int),
                ('serial', c_ulong),
                ('send_event', Bool),
                ('display', POINTER(Display)),
                ('window', Window),
                ('root', Window),
                ('subwindow', Window),
                ('time', Time),
                ('x', c_int),
                ('y', c_int),
                ('x_root', c_int),
                ('y_root', c_int),
                ('state', c_uint),
                ('button', c_uint),
                ('same_screen', Bool)]

class XMotionEvent(Structure):
    _fields_ = [('type', c_int),
                ('serial', c_ulong),
                ('send_event', Bool),
                ('display', POINTER(Display)),
                ('window', Window),
                ('root', Window),
                ('subwindow', Window),
                ('time', Time),
                ('x', c_int),
                ('y', c_int),
                ('x_root', c_int),
                ('y_root', c_int),
                ('state', c_uint),
                ('is_hint', c_ubyte),
                ('same_screen', Bool)]

class XClientMessageEvent(Structure):
    _fields_ = [('type', c_int),
                ('serial', c_ulong),
                ('send_event', Bool),
                ('display', POINTER(Display)),
                ('window', Window),
                ('message_type', Atom),
                ('format', c_int),
                ('data', c_long * 5)]

class XEvent(Union):
    _fields_ = [('type', c_int),
                ('xkey', XKeyEvent),
                ('xbutton', XButtonEvent),
                ('xmotion', XMotionEvent),
                ('xclient', XClientMessageEvent),
                ('pad', c_long * 24)]

class XSetWindowAttributes(Structure):
    _fields_ = [('background_pixmap', Pixmap),
                ('background_pixel', c_ulong),
                ('border_pixmap', Pixmap),
                ('border_pixel', c_ulong),
                ('bit_gravity', c_int),
                ('win_gravity', c_int),
                ('backing_store', c_int),
                ('backing_planes', c_ulong),
                ('backing_pixel', c_ulong),
                ('save_under', Bool),
                ('event_mask', c_long),
                ('do_not_propagate_mask', c_long),
                ('override_redirect', Bool),
                ('colormap', Colormap),
                ('cursor', Cursor)]


# constants
KeyPress = 2
KeyRelease = 3
ButtonPress = 4
ButtonRelease = 5
MotionNotify = 6
FocusIn = 9
FocusOut = 10
ClientMessage = 33

CopyFromParent = 0
InputOnly = 2

CWOverrideRedirect = (1<<9)

ShiftMask = (1<<0)
LockMask = (1<<1)
ControlMask = (1<<2)
Mod1Mask = (1<<3)
Mod2Mask = (1<<4)
Mod3Mask = (1<<5)
Mod4Mask = (1<<6)
Mod5Mask = (1<<7)


# functions
XOpenDisplay = libX11.XOpenDisplay
XOpenDisplay.argtypes = [String]
XOpenDisplay.restype = POINTER(Display)

XCloseDisplay = libX11.XCloseDisplay
XCloseDisplay.argtypes = [POINTER(Display)]
XCloseDisplay.restype = c_int

XConnectionNumber = libX11.XConnectionNumber
XConnectionNumber.argtypes = [POINTER(Display)]
XConnectionNumber.restype = c_int

XInternAtom = libX11.XInternAtom
XInternAtom.argtypes = [POINTER(Display), String, Bool]
XInternAtom.restype = c_int

XDefaultRootWindow = libX11.XDefaultRootWindow
XDefaultRootWindow.argtypes = [POINTER(Display)]
XDefaultRootWindow.restype = Window

XCreateWindow = libX11.XCreateWindow
XCreateWindow.argtypes = [POINTER(Display), Window, c_int, c_int, c_uint, c_uint, c_uint, c_int, c_uint, POINTER(Visual), c_ulong, POINTER(XSetWindowAttributes)]
XCreateWindow.restype = Window

XDestroyWindow = libX11.XDestroyWindow
XDestroyWindow.argtypes = [POINTER(Display), Window]
XDestroyWindow.restype = c_int

XFree = libX11.XFree
XFree.argtypes = [POINTER(None)]
XFree.restype = c_int

XNextEvent = libX11.XNextEvent
XNextEvent.argtypes = [POINTER(Display), POINTER(XEvent)]
XNextEvent.restype = c_int

XPeekEvent = libX11.XPeekEvent
XPeekEvent.argtypes = [POINTER(Display), POINTER(XEvent)]
XPeekEvent.restype = c_int

XSendEvent = libX11.XSendEvent
XSendEvent.argtypes = [POINTER(Display), Window, c_int, c_long, POINTER(XEvent)]
XSendEvent.restype = c_int

XFlush = libX11.XFlush
XFlush.argtypes = [POINTER(Display)]
XFlush.restype = c_int

XPending = libX11.XPending
XPending.argtypes = [POINTER(Display)]
XPending.restype = c_int

XSynchronize = libX11.XSynchronize
XSynchronize.argtypes = [POINTER(Display), c_int]
XSynchronize.restype = POINTER(CFUNCTYPE(c_int, POINTER(Display)))


## xim
# types
class _XIC(Structure):
    pass

XIC = POINTER(_XIC)

class _XIM(Structure):
    pass

XIM = POINTER(_XIM)

class _XrmDatabase(Structure):
    pass

XrmDatabase = POINTER(_XrmDatabase)


# constants
XNInputStyle = b'inputStyle'
XNClientWindow = b'clientWindow'

XIMPreeditNothing = 0x0008
XIMPreeditNone = 0x0010
XIMStatusNothing = 0x0400
XIMStatusNone = 0x0800

XBufferOverflow = -1
NoSymbol = 0
XLookupNone = 1
XLookupChars = 2
XLookupKeySym = 3
XLookupBoth = 4


# functions
XFilterEvent = libX11.XFilterEvent
XFilterEvent.argtypes = [POINTER(XEvent), Window]
XFilterEvent.restype = c_int

XOpenIM = libX11.XOpenIM
XOpenIM.argtypes = [POINTER(Display), XrmDatabase, String, String]
XOpenIM.restype = XIM

XCloseIM = libX11.XCloseIM
XCloseIM.argtypes = [XIM]
XCloseIM.restype = c_int

XCreateIC = libX11.XCreateIC
XCreateIC.restype = XIC

XDestroyIC = libX11.XDestroyIC
XDestroyIC.argtypes = [XIC]
XDestroyIC.restype = None

XSetICFocus = libX11.XSetICFocus
XSetICFocus.argtypes = [XIC]
XSetICFocus.restype = None

Xutf8ResetIC = libX11.Xutf8ResetIC
Xutf8ResetIC.argtypes = [XIC]
Xutf8ResetIC.restype = String

Xutf8LookupString = libX11.Xutf8LookupString
Xutf8LookupString.argtypes = [XIC, POINTER(XKeyPressedEvent), String, c_int, POINTER(KeySym), POINTER(c_int)]
Xutf8LookupString.restype = c_int

XKeysymToString = libX11.XKeysymToString
XKeysymToString.argtypes = [KeySym]
XKeysymToString.restype = String

XkbKeycodeToKeysym = libX11.XkbKeycodeToKeysym
XkbKeycodeToKeysym.argtypes = [POINTER(Display), KeyCode, c_uint, c_uint]
XkbKeycodeToKeysym.restype = KeySym


## record extensions
libXtst = CDLL('libXtst.so')

# types
XPointer = String
XRecordContext = c_ulong
XRecordClientSpec = c_ulong

class XRecordRange8(Structure):
    _fields_ = [('first', c_ubyte),
                ('last', c_ubyte)]

class XRecordRange16(Structure):
    _fields_ = [('first', c_ushort),
                ('last', c_ushort)]

class XRecordExtRange(Structure):
    _fields_ = [('ext_major', XRecordRange8),
                ('ext_minor', XRecordRange16)]

class XRecordRange(Structure):
    _fields_ = [('core_requests', XRecordRange8),
                ('core_replies', XRecordRange8),
                ('ext_requests', XRecordExtRange),
                ('ext_replies', XRecordExtRange),
                ('delivered_events', XRecordRange8),
                ('device_events', XRecordRange8),
                ('errors', XRecordRange8),
                ('client_started', c_int),
                ('client_died', c_int)]

class XRecordInterceptData(Structure):
    _fields_ = [('id_base', XID),
                ('server_time', Time),
                ('client_seq', c_ulong),
                ('category', c_int),
                ('client_swapped', c_int),
                ('data', POINTER(c_ubyte)),
                ('data_len', c_ulong)]

XRecordInterceptProc = CFUNCTYPE(None, XPointer, POINTER(XRecordInterceptData))


# constants
XRecordAllClients = 3
XRecordFromServer = 0

# functions
XRecordAllocRange = libXtst.XRecordAllocRange
XRecordAllocRange.argtypes = []
XRecordAllocRange.restype = POINTER(XRecordRange)

XRecordCreateContext = libXtst.XRecordCreateContext
XRecordCreateContext.argtypes = [POINTER(Display), c_int, POINTER(XRecordClientSpec), c_int, POINTER(POINTER(XRecordRange)), c_int]
XRecordCreateContext.restype = XRecordContext

XRecordEnableContextAsync = libXtst.XRecordEnableContextAsync
XRecordEnableContextAsync.argtypes = [POINTER(Display), XRecordContext, XRecordInterceptProc, XPointer]
XRecordEnableContextAsync.restype = c_int

XRecordProcessReplies = libXtst.XRecordProcessReplies
XRecordProcessReplies.argtypes = [POINTER(Display)]
XRecordProcessReplies.restype = None

XRecordDisableContext = libXtst.XRecordDisableContext
XRecordDisableContext.argtypes = [POINTER(Display), XRecordContext]
XRecordDisableContext.restype = c_int

XRecordFreeContext = libXtst.XRecordFreeContext
XRecordFreeContext.argtypes = [POINTER(Display), XRecordContext]
XRecordFreeContext.restype = c_int

XRecordFreeData = libXtst.XRecordFreeData
XRecordFreeData.argtypes = [POINTER(XRecordInterceptData)]
XRecordFreeData.restype = None


## wire protocol
CARD8 = c_ubyte
CARD16 = c_ushort
CARD32 = c_uint
BOOL = CARD8
BYTE = CARD8
INT16 = c_short

class xEventType(Structure):
    _fields_ = [('type', BYTE),
                ('detail', BYTE),
                ('sequenceNumber', CARD16)]

class xKeyButtonPointer(Structure):
    _fields_ = [('pad00', CARD32),
                ('time', CARD32),
                ('root', CARD32),
                ('event', CARD32),
                ('child', CARD32),
                ('rootX', INT16),
                ('rootY', INT16),
                ('eventX', INT16),
                ('eventY', INT16),
                ('state', CARD16),
                ('sameScreen', BOOL),
                ('pad1', BYTE)]

class xEvent(Union):
    _fields_ = [('u', xEventType),
                ('keyButtonPointer', xKeyButtonPointer)]


def _kbd_wire_to_event(dpy, wev):
    ev = XEvent()
    ev.xkey.type = wev.u.type
    ev.xkey.serial = wev.u.sequenceNumber
    ev.xkey.send_event = ((wev.u.type & 0x80) != 0)
    ev.xkey.display = dpy
    ev.xkey.window = wev.keyButtonPointer.event
    ev.xkey.root = wev.keyButtonPointer.root
    ev.xkey.subwindow = wev.keyButtonPointer.child
    ev.xkey.time = wev.keyButtonPointer.time
    ev.xkey.x = wev.keyButtonPointer.eventX
    ev.xkey.y = wev.keyButtonPointer.eventY
    ev.xkey.x_root = wev.keyButtonPointer.rootX
    ev.xkey.y_root = wev.keyButtonPointer.rootY
    ev.xkey.state = wev.keyButtonPointer.state
    ev.xkey.keycode = wev.u.detail
    ev.xkey.same_screen = wev.keyButtonPointer.sameScreen
    return ev


def _btn_wire_to_event(dpy, wev):
    ev = XEvent()
    ev.xbutton.type = wev.u.type
    ev.xbutton.serial = wev.u.sequenceNumber
    ev.xbutton.send_event = ((wev.u.type & 0x80) != 0)
    ev.xbutton.display = dpy
    ev.xbutton.window = wev.keyButtonPointer.event
    ev.xbutton.root = wev.keyButtonPointer.root
    ev.xbutton.subwindow = wev.keyButtonPointer.child
    ev.xbutton.time = wev.keyButtonPointer.time
    ev.xbutton.x = wev.keyButtonPointer.eventX
    ev.xbutton.y = wev.keyButtonPointer.eventY
    ev.xbutton.x_root = wev.keyButtonPointer.rootX
    ev.xbutton.y_root = wev.keyButtonPointer.rootY
    ev.xbutton.state = wev.keyButtonPointer.state
    ev.xbutton.button = wev.u.detail
    ev.xbutton.same_screen = wev.keyButtonPointer.sameScreen
    return ev


def _mtn_wire_to_event(dpy, wev):
    ev = XEvent()
    ev.xmotion.type = wev.u.type
    ev.xmotion.serial = wev.u.sequenceNumber
    ev.xmotion.send_event = ((wev.u.type & 0x80) != 0)
    ev.xmotion.display = dpy
    ev.xmotion.window = wev.keyButtonPointer.event
    ev.xmotion.root = wev.keyButtonPointer.root
    ev.xmotion.subwindow = wev.keyButtonPointer.child
    ev.xmotion.time = wev.keyButtonPointer.time
    ev.xmotion.x = wev.keyButtonPointer.eventX
    ev.xmotion.y = wev.keyButtonPointer.eventY
    ev.xmotion.x_root = wev.keyButtonPointer.rootX
    ev.xmotion.y_root = wev.keyButtonPointer.rootY
    ev.xmotion.state = wev.keyButtonPointer.state
    ev.xmotion.is_hint = wev.u.detail
    ev.xmotion.same_screen = wev.keyButtonPointer.sameScreen
    return ev


def XWireToEvent(dpy, data):
    # this could have been avoided if _XWireToEvent didn't have internal state
    wev = cast(data, POINTER(xEvent)).contents
    if wev.u.type in [KeyPress, KeyRelease]:
        return _kbd_wire_to_event(dpy, wev)
    elif wev.u.type in [ButtonPress, ButtonRelease]:
        return _btn_wire_to_event(dpy, wev)
    elif wev.u.type == MotionNotify:
        return _mtn_wire_to_event(dpy, wev)
    return XEvent(wev.u.type)
